# tag: table rendering base class (backup)

# TODO:
# - field attribute row_align support
# - limit support

package HTML::Tabulate;

use 5.005;
use Carp;
use Data::Dumper;
use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(&render);

$VERSION = '0.09';
my %VALID_ARG = (
    table => 'HASH/SCALAR',
    tr => 'HASH', 
    thtr => 'HASH',
    th => 'HASH',
    td => 'HASH',
    fields => 'ARRAY',
    in_fields => 'ARRAY',
    labels => 'SCALAR/HASH',
    label_links => 'HASH',
    stripe => 'ARRAY/SCALAR',
    null => 'SCALAR',
    trim => 'SCALAR',
    style => 'SCALAR',
#   limit => 'SCALAR',
#   output => 'SCALAR',
#   first => 'SCALAR',
#   last => 'SCALAR',
    field_attr => 'HASH',
);
my %VALID_FIELDS = (
    -defaults => 'HASH',
);
my %FIELD_ATTR = (
    escape => 'SCALAR',
    value => 'SCALAR/CODE/ARRAY',
    format => 'SCALAR/CODE',
    link => 'SCALAR/CODE',
    label => 'SCALAR/ARRAY',
    label_link => 'SCALAR/CODE',
);

# -----------------------------------------------------------------------------------------
# Provided for subclassing
sub get_valid_arg
{
    return wantarray ? %VALID_ARG : \%VALID_ARG;
}

# Provided for subclassing
sub get_valid_fields
{
    return wantarray ? %VALID_FIELDS : \%VALID_FIELDS;
}

# Provided for subclassing
sub get_field_attributes
{
    return wantarray ? %FIELD_ATTR : \%FIELD_ATTR;
}

#
# Check $self->{defn} for invalid arguments or types
#
sub check_valid
{
    my ($self, $defn) = @_;

    # Check top-level args
    my %valid = $self->get_valid_arg();
    my (@invalid, @badtype);
    for (sort keys %$defn) {
        push @invalid, $_ if ! exists $valid{$_};
        my $type = ref $defn->{$_};
        push @badtype, $_ 
            if $type && $type ne 'SCALAR' && $valid{$_} !~ m/$type/;
        push @badtype, $_ 
            if ! $type && $valid{$_} !~ m/SCALAR/;
    }
    croak "[check_valid] invalid argument found: " . join(',',@invalid) 
        if @invalid;
    croak "[check_valid] invalid types for argument: " . join(',',@badtype)
        if @badtype;

    # Check special fields
    %valid = $self->get_valid_fields();
    @invalid = ();
    @badtype = ();
    for (sort grep(/^-/, keys(%{$defn->{field}})) ) {
        push @invalid, $_ if ! exists $valid{$_};
        my $type = ref $defn->{field}->{$_};
        push @badtype, $_ 
            if $type && $type ne 'SCALAR' && $valid{$_} !~ m/$type/;
        push @badtype, $_ 
            if ! $type && $valid{$_} !~ m/SCALAR/;
    }
    croak "[check_valid] invalid field argument found: " . join(',',@invalid) 
        if @invalid;
    croak "[check_valid] invalid types for field argument: " . join(',',@badtype)
        if @badtype;

    # Check field attributes
    $self->{field_attr} ||= $self->get_field_attributes();
    %valid = %{$self->{field_attr}};
    @badtype = ();
    for my $field (keys %{$defn->{field}}) {
        croak "[check_valid] invalid field argument entry '$field': " . 
            $defn->{field}->{$field} if ref $defn->{field}->{$field} ne 'HASH';
        for (sort keys %{$defn->{field}->{$field}}) {
            next if ! exists $valid{$_};
            next if ! $valid{$_};
            my $type = ref $defn->{field}->{$field}->{$_};
            if (! ref $valid{$_}) {
                push @badtype, $_ 
                    if $type && $type ne 'SCALAR' && $valid{$_} !~ m/$type/;
                push @badtype, $_ 
                    if ! $type && $valid{$_} !~ m/SCALAR/;
            }
            elsif (ref $valid{$_} eq 'ARRAY') {
                if ($type) {
                    push @badtype, $_;
                }
                else {
                    my $val = $defn->{field}->{$field}->{$_};
                    push @badtype, "$_ ($val)" if ! grep /^$val$/, @{$valid{$_}};
                }
            }
            else {
                croak "[check_valid] invalid field attribute entry for '$_': " . 
                    ref $valid{$_};
            }
        }
        croak "[check_valid] invalid type for '$field' field attribute: " . 
            join(',',@badtype) if @badtype;
    }
}

#
# Merge $hash1 and $hash2 together, returning the result. Performs a 
#   shallow (one-level deep) hash merge unless the field is defined in 
#   the @recurse_keys array, in which case we do a full recursive merge.
#
sub merge
{
    my $self = shift;
    my $hash1 = shift || {};
    my $hash2 = shift;

    croak "[merge] invalid hash1 '$hash1'" if ref $hash1 ne 'HASH';
    croak "[merge] invalid hash2 '$hash2'" if $hash2 && ref $hash2 ne 'HASH';

    my $single_arg = ! $hash2;

    # Use $self->{defn} as $hash1 if only one argument
    if ($single_arg) {
        $hash2 = $hash1;
        $hash1 = $self->{defn};
    }

    # Check hash2 for valid args (except when recursive)
    my $sub = (caller(1))[3];
    $self->check_valid($hash2) if substr($sub, -7) ne '::merge';

    my %merge;
    # Add hash1 to %merge
    @merge{ keys %$hash1 } = values %$hash1;

    # Add hash2 to %merge 
    my @recurse_keys = qw(field);
    for my $key (keys %$hash2) {
        # If this value is a hashref on both sides, do a shallow hash merge
        #   unless we need to do a proper recursive merge
        if (ref $hash2->{$key} eq 'HASH' && ref $merge{$key} eq 'HASH') {
            # Recursive merge
            if (grep /^$key$/, @recurse_keys) {
                $merge{$key} = $self->merge($hash1->{$key}, $hash2->{$key});
            }
            # Shallow hash merge
            else {
                @{$merge{$key}}{ keys %{$hash1->{$key}}, keys %{$hash2->{$key}} } = (values %{$hash1->{$key}}, values %{$hash2->{$key}});
            }
        }
        # Otherwise (scalars, arrayrefs etc) just copy the value
        else {
            $merge{$key} = $hash2->{$key};
        }
    }

    # In void context update $self->{defn} in single_arg mode, else $hash2
    if (! defined wantarray) {
#       if ($single_arg) {
            $self->{defn} = \%merge;
#       }
#       else {
#           @$hash2{ keys %merge } = values %merge;
#       }
    }
    else {
        return \%merge 
    }
}

# Initialisation
sub init
{
    my $self = shift;
    my $defn = shift || {};
    croak "[init] invalid defn '$defn'" if $defn && ref $defn ne 'HASH';

    # Map $defn table => 1 to table => {} for cleaner merging
    $defn->{table} = {} if $defn->{table} && ! ref $defn->{table};

    # Initialise $self->{defn} by merging defaults and $defn
    $self->{defn} = $self->merge({
        style => 'down', 
        table => {},
        field => { -defaults => {}, },
    }, $defn);

    return $self;
}

sub new 
{
    my $class = shift;
    my $self = {};
    bless $self, $class;
    $self->init(@_);
}

# -----------------------------------------------------------------------------------------
#
# Try and derive a reasonable field list using the set data.
#   Croaks on failure.
#
sub derive_fields
{
    my ($self, $set) = @_;
    my $defn = $self->{tmp};

    # For DBIx::Recordset object, prefetch the first row
    if (ref $set eq 'DBIx::Recordset') {
        require DBIx::Recordset;
        my $row = $set->First();
        $self->{prefetch} = $row;
        $defn->{fields} = [ sort keys %$row ] if ref $row;
    }
    # For arrays
    elsif (ref $set eq 'ARRAY' && @$set) {
        # Arrayref of hashrefs
        if (ref $set->[0] eq 'HASH') {
            $defn->{fields} = [ sort keys %{$set->[0]} ];
        }
        elsif (ref $set->[0] ne 'ARRAY') {
            croak "[render] no fields found and first row '" . $set->[0] . "' is strange type";
        }
        # Arrayref of arrayrefs - fatal only if labels => 1
        elsif ($defn->{labels}) {
            croak "[render] no fields found and cannot derive fields from data arrayrefs";
        }
        # Arrayref of arrayrefs, labels off
        else {
            $defn->{fields} = [ 0 .. $#{$set->[0]} ];
        }
    }
    # Check for hashref (possibly blessed)
    else {
        if (defined eval { keys %$set }) {
            my $first = (sort keys %$set)[0];
            # Check whether first value is reference
            if (my $ref = ref $set->{$first}) {
                # Hashref of hashrefs
                if ($ref eq 'HASH') {
                    $defn->{fields} = [ sort keys %{$set->{$first}} ];
                }
                elsif (ref $set->[0] ne 'ARRAY') {
                    croak "[render] no fields found and first row '" . $set->[0] . "' is strange type";
                }
                # Hashref of arrayrefs - fatal only if labels => 1
                elsif ($defn->{labels}) {
                    croak "[render] no fields found and cannot derive fields from data arrayrefs";
                }
                # Hashref of arrayrefs, labels off
                else {
                    $defn->{fields} = [ 0 .. $#{$set->[$first]} ];
                }
            }
            else {
                $defn->{fields} = [ sort keys %$set ];
            }
        }
        else {
            croak "[render] no fields found and set '$set' is strange type: $@";
        }
    }
}

#
# Deep copy routine, swiped from a Randal Schwartz column
#
sub deepcopy 
{
    my $this = shift;
    if (not ref $this) {
        $this;
    } elsif (ref $this eq "ARRAY") {
        [map deepcopy($_), @$this];
    } elsif (ref $this eq "HASH") {
        {map { $_ => deepcopy($this->{$_}) } keys %$this};
    } else {
        die "what type is $_?"
    }
}

#
# Create a derived table definition by doing a set of one-off or 
#   dataset-specific mappings on the current table definition e.g. deriving 
#   a field list if none is set, setting up a field map for arrayref-of-
#   arrayref sets, and mapping top-level shortcuts into their field 
#   attribute equivalents.
#
sub munge_defn
{
    my ($self, $set) = @_;

    # If already done, return unless we require any dataset-specific mappings
    if ($self->{tmp}) {
        return unless 
            ref $self->{defn}->{fields} ne 'ARRAY' || 
            ! @{$self->{defn}->{fields}} ||
            (ref $set eq 'ARRAY' && @$set && ref $set->[0] eq 'ARRAY');
    }

    # Copy $self->{defn} to $self->{tmp}
    $self->{tmp} = deepcopy($self->{defn});
    my $defn = $self->{tmp};

    # Try to derive field list if not set
    $self->derive_fields($set) 
        if ref $defn->{fields} ne 'ARRAY' || ! @{$defn->{fields}};

    # Set up a field map for arrayref-of-arrayref sets
    my $pos = 0;
    my $fields = ref $defn->{in_fields} eq 'ARRAY' ? $defn->{in_fields} : $defn->{fields};
    $defn->{field_map} = { map  { $_ => $pos++; } @$fields }
        if ref $set eq 'ARRAY' && @$set && ref $set->[0] eq 'ARRAY';

    # Map top-level 'labels' and 'label_links' hashrefs into fields
    if (ref $defn->{labels} eq 'HASH') {
        for (keys %{$defn->{labels}}) {
            $defn->{field}->{$_} ||= {};
            $defn->{field}->{$_}->{label} = $defn->{labels}->{$_};
        }
    }
    if (ref $defn->{label_links} eq 'HASH') {
        for (keys %{$defn->{label_links}}) {
            $defn->{field}->{$_} ||= {};
            $defn->{field}->{$_}->{label_link} = $defn->{label_links}->{$_};
        }
    }

    # Map top-level 'th' hashref into -defaults label_ attributes
    if (ref $defn->{th} eq 'HASH') {
        for (keys %{$defn->{th}}) {
            $defn->{field}->{-defaults}->{"label_$_"} = $defn->{th}->{$_}
                if ! exists $defn->{field}->{-defaults}->{"label_$_"};
        }
    }
    # Map top-level 'td' hashref into -defaults
    if (ref $defn->{td} eq 'HASH') {
        $defn->{field}->{-defaults} = { %{$defn->{td}}, %{$defn->{field}->{-defaults}} };
    }
}

# -----------------------------------------------------------------------------------------
#
# Return the given HTML $tag with attributes from the $attr hashref.
#   An attribute with a non-empty value (i.e. not '' or undef) is rendered
#   attr="value"; one with a value of '' is rendered as a 'bare' attribute
#   (i.e. no '='); one with undef is ignored (allowing unset CGI parameters
#   to be ignored).
#
sub start_tag
{
    my ($self, $tag, $attr) = @_;
    my $str = "<$tag";
    if (ref $attr eq 'HASH') {
        for my $a (sort keys %$attr) {
            if ($attr->{$a} ne '') {
                $str .= qq( $a="$attr->{$a}");
            }
            elsif (defined $attr->{$a}) {
                $str .= qq( $a);
            }
        }
    }
    $str .= ">";
    return $str;
}

sub end_tag
{
    my ($self, $tag) = @_;
    return "</$tag>";
}

# Provided for subclassing
sub start_table 
{
    my ($self) = @_;
    return '' if exists $self->{defn}->{table} && ! $self->{defn}->{table};
    return $self->start_tag('table',$self->{defn}->{table}) . "\n";
}

# Provided for subclassing
sub end_table 
{
    my ($self) = @_;
    return '' if exists $self->{defn}->{table} && ! $self->{defn}->{table};
    return $self->end_tag('table') . "\n";
}

#
# Format the given data item using formatting field attributes (e.g. format, 
#   link, escape etc.)
#
sub cell_format
{
    my ($self, $data, $fattr, $row, $field) = @_;
    my $data_in = $data;

    # 'format' subroutine or sprintf pattern
    if ($fattr->{format}) {
        my $ref = ref $fattr->{format};
        croak "[cell_format] invalid '$field' format: $ref"
            if $ref && $ref ne 'CODE';
        $data = &{$fattr->{format}}($data, $row || {}) if $ref eq 'CODE';
        $data = sprintf $fattr->{format}, $data if ! $ref;
    }

    # 'escape' boolean for simple tag escaping (defaults to on)
    if ($fattr->{escape} || ! exists $fattr->{escape}) {
        $data =~ s/</&lt;/g;
        $data =~ s/>/&gt;/g;
    }

    # 'link' subroutine or sprintf pattern
    if ($fattr->{link}) {
        my $ldata;
        my $ref = ref $fattr->{link};
        croak "[cell_format] invalid '$field' link: $ref"
            if $ref && $ref ne 'CODE';
        $ldata = &{$fattr->{link}}($data_in, $row || {}) if $ref eq 'CODE';
        $ldata = sprintf $fattr->{link}, $data_in if ! $ref;
        $data = qq(<a href="$ldata">$data</a>);
    }

    return $data;
}

sub label
{
    my ($self, $field) = @_;
    return ucfirst($field) if ! ref $self->{defn}->{field}->{$field};

    # Use first label if arrayref
    if (ref $self->{defn}->{field}->{$field}->{label} eq 'ARRAY') {
        return $self->{defn}->{field}->{$field}->{label}->[0];
    }
    else {
        return $self->{defn}->{field}->{$field}->{label} || ucfirst($field);
    }
}

#
# Add in any extra (conditional) defaults for this field. 
# Provided for subclassing.
#
sub cell_merge_extras
{
    return ();
}

#
# Merge default and field attributes once each for labels and data
#
sub cell_merge_defaults
{
    my ($self, $row, $field) = @_;
    $self->{defn}->{field}->{$field} ||= {};
    my $fattr = { %{$self->{defn}->{field}->{-defaults}},
                  $self->cell_merge_extras($row, $field),
                  %{$self->{defn}->{field}->{$field}}, };

    # For labels, merge and defaults + field attributes, keeping only label_ ones
    if ($row == 0) {
        my $fattr2 = {};
        # Copy and convert label_ attributes into $fattr
        for (keys %$fattr) {
            next if substr($_,0,6) ne 'label_';
            $fattr2->{substr($_,6)} = $fattr->{$_};
        }
        # Set 'value' from 'label'
        $fattr2->{value} = $self->label($field);
        $fattr = $fattr2;
    }

    # For data, merge defaults and the field attributes, discarding label_ ones
    else {
        # Remove all label_ attributes
        for (keys %$fattr) { delete $fattr->{$_} if substr($_,0,6) eq 'label_'; }
    }

    # Create td_attr by removing all $fattr attributes in $field_attr
    my %td_attr = %$fattr;
    for (keys %td_attr) { delete $td_attr{$_} if exists $self->{field_attr}->{$_} }

    # If data, save for subsequent rows
    if ($row) {
        $fattr->{td_attr} = \%td_attr;
        $self->{defn}->{field}->{$field} = $fattr;
    }

    return ($fattr, \%td_attr);
}

#
# Set and format the data for a single (data) cell or item
#
sub cell_content
{
    my ($self, $row, $field, $fattr) = @_;
    $fattr ||= $self->{defn}->{field}->{$field};

    # Set data from $row
    my $data;
    if (ref $row eq 'ARRAY') {
        my $i = keys %{$self->{defn}->{field_map}} ? $self->{defn}->{field_map}->{$field} : $field;
        $data = $row->[ $i ];
    }
    elsif (ref $row eq 'HASH') {
        $data = $row->{$field};
    }
    # 'value' literal or subref takes precedence over row
    if ($fattr->{value}) {
        my $ref = ref $fattr->{value};
        if (! $ref) {
            # $data = sprintf $fattr->{value}, $data;
            $data = $fattr->{value};
        }
        elsif ($ref eq 'CODE') {
            $data = &{$fattr->{value}}($data, $row);
        }
        elsif ($ref eq 'ARRAY') {
            $data = $fattr->{value}->[0];
        }
        else {
            croak "[cell_content] invalid '$field' value: $ref";
        };
    }

    # Format
    $data =~ s/^\s*(.*?)\s*$/$1/ if $self->{defn}->{trim};
    $data = $self->cell_format($data, $fattr, $row, $field);
    $data = $self->{defn}->{null} if defined $self->{defn}->{null} && $data eq '';

    return $data;
}

#
# Wrap cell in <th> or <td> table tags
#
sub cell_tags 
{
    my ($self, $data, $row, $field, $tx_attr) = @_;

    my $tag = ($row == 0) ? 'th' : 'td';
    return $self->start_tag($tag, $tx_attr) . $data . $self->end_tag($tag);
}

#
# Render a single table cell or item
#
sub cell 
{
    my ($self, $row, $field) = @_;
    my ($fattr, $tx_attr);

    # Merge default and field attributes first time through (labels + data)
    if ($row == 0 || ! $self->{defn}->{field}->{$field}->{td_attr}) {
        ($fattr, $tx_attr) = $self->cell_merge_defaults($row, $field);
    }
    else {
        $tx_attr = $self->{defn}->{field}->{$field}->{td_attr};
    }

    # Generate output
    my $out = $self->cell_content($row, $field, $fattr);
    return $self->cell_tags($out, $row, $field, $tx_attr);
}

#
# Modify the $h hashref with a bgcolor attribute for every nth row.
#   $stripe is either an n-ary arrayref of colour values or a scalar 
#   (a scalar is used as a binary stripe)
#
sub stripe
{
    my ($self, $h, $rownum, $stripe) = @_;
    return if $stripe eq '' || ref $h ne 'HASH';
             
    # Treat a scalar stripe as the second of a binary stripe
    $stripe = [ undef, $stripe ] if ! ref $stripe;
                     
    my $r = int($rownum % scalar(@$stripe)) - 1;
    if ($stripe->[$r]) {
        $h->{bgcolor} = $stripe->[$r]
            unless $rownum == 0 && exists $h->{bgcolor};
    }
    else {
        delete $h->{bgcolor} unless $rownum == 0;
    }
}

#
# Render a single table row (style 'down')
#
sub row_down 
{
    my ($self, $row, $rownum) = @_;
    my $defn = $self->{defn};

    # Table striping
    $defn->{tr} = {} if ref $defn->{tr} ne 'HASH';
    my $tr;
    if ($rownum == 0) {
        $tr = $defn->{thtr} || $defn->{tr};
    }
    else {
        $tr = $defn->{tr};
    }
    $self->stripe($tr, $rownum, $defn->{stripe}) if $defn->{stripe};

    # Render row into $out
    my $out = '';
    $out .= $self->start_tag('tr', $tr);
    for my $f (@{$defn->{fields}}) {
        # Labels/headings
        if ($rownum == 0) {
            $out .= $self->cell(0, $f);
        }
        # Data
        else {
            $out .= $self->cell($row, $f);
        }
    }
    $out .= $self->end_tag('tr');
    $out .= "\n";
}

#
# Render the table body with successive records down the page
#
sub body_down 
{
    my ($self, $set) = @_;
    my $body = '';

    # Labels/headings
    my @fields = @{$self->{defn}->{fields}} 
        if ref $self->{defn}->{fields} eq 'ARRAY';
    if ($self->{defn}->{labels} && @fields) {
        $body .= $self->row_down(undef, 0);
    }

    # Data
    my $r = 1;
    my $row = $self->{prefetch};
    if (ref $set eq 'DBIx::Recordset') {
        require DBIx::Recordset;
        $row ||= $set->First();
        do {
            $body .= $self->row_down($row, $r);
            $r++;
        } while ($row = $set->Next());
    }
    elsif (ref $set eq 'ARRAY') {
        for my $row (@$set) {
            $body .= $self->row_down($row, $r);
            $r++;
        }
    }
    elsif (ref $set eq 'HASH' || eval { keys %$set }) {
        # Check first value - drill down further unless non-reference
        my $k = (sort keys %$set)[0];
        if (! ref $set->{$k}) {
            $body .= $self->row_down($set, $r);
            $r++;
        }
        else {
            for $k (sort keys %$set) {
                $body .= $self->row_down($set->{$k}, $r);
                $r++;
            }
        }
    }
    else {
        croak "invalid Tabulate data type '$set'";
    }

    return $body;
}

#
# Render a single table row (style 'across')
#
sub row_across
{
    my ($self, $data, $rownum, $field) = @_;
    my $tr = $self->{defn}->{tr} || {};

    # Begin row
    $self->stripe($tr, $rownum, $self->{defn}->{stripe}) 
        if $self->{defn}->{stripe};
    my $body = $self->start_tag('tr', $tr);

    # Label/heading
    $body .= $self->cell(0, $field) if $self->{defn}->{labels};

    # Data
    for my $row (@$data) {
        $body .= $self->cell($row, $field);
    }

    # End row
    $rownum++;
    $body .= $self->end_tag('tr', $tr) . "\n";
}

#
# Render the table body with successive records across the page 
#   (i.e. fields down the page)
#
sub body_across 
{
    my ($self, $set) = @_;

    # Fetch the full data set
    my @data = ();
    if (ref $set eq 'DBIx::Recordset') {
        require DBIx::Recordset;
        my $row = $set->First();
        if (ref $row) {
            do {
                push @data, $row;
            }
            while ($row = $set->Next());
        }
    }
    elsif (ref $set eq 'ARRAY') {
        @data = @$set;
    }
    elsif (ref $set eq 'HASH' || eval { keys %$set }) {
        @data = ( $set );
    }
    else {
        croak "invalid Tabulate data type '$set'";
    }

    # Iterate over fields
    my $rownum = 1;
    my $body = '';
    for my $field (@{$self->{defn}->{fields}}) {
        $body .= $self->row_across(\@data, $rownum, $field);
    }

    return $body;
}

# -----------------------------------------------------------------------------------------
#
# Render the data set $set using the settings in $self->{defn} + $defn,
#   returning the resulting string.
#
sub render
{
    my ($self, $set, $defn) = @_;
    croak "[render] invalid set '$set'" if ! ref $set;

    # If $self not blessed, this is a procedural call, $self is $defn
    $self = HTML::Tabulate->new($self) if ref $self eq 'HASH';

    # Ensure the definition set has been appropriately munged
    $self->munge_defn($set);

    # Start table
    my $out = $self->start_table();

    # Style-specific bodies (default is 'down')
    if ($self->{defn}->{style} eq 'down') {
        $out .= $self->body_down($set);
    }
    elsif ($self->{defn}->{style} eq 'across') {
        $out .= $self->body_across($set);
    }
    else {
        croak "[render] invalid style '$self->{defn}->{style}'";
    }

    # End table
    $out .= $self->end_table();

    return $out;
}

1;

__END__

=head1 NAME

HTML::Tabulate - HTML table rendering class


=head1 SYNOPSIS

    use HTML::Tabulate;

    # Setup initial defaults
    $t = HTML::Tabulate->new({ 
        table => { border => 0, cellpadding => 0, cellspacing => 3 },
    });

    # Setup some data-specific settings
    $table = {
        fields => [ qw(emp_id name title) ],
        labels => 1,
        field => {
            emp_id => {
                format => '%-05d',
                link => "emp.html?id=%s",
                align => 'right',
            },
            name => {
                format => sub { uc(shift) },
            }
        },
    };

    # Render the data
    print $t->render($data, $table);


    
=head1 DESCRIPTION

HTML::Tabulate is used to render/display a given set of data in 
an HTML table. It takes a data set and a presentation definition 
and applies the presentation to the data set to produce the HTML 
table output. The presentation definition accepts arguments 
corresponding to HTML table tags ('table', 'tr', 'th', 'td' etc.)
to define attributes for those tags, plus additional arguments 
for other aspects of the presentation. HTML::Tabulate supports 
advanced features like automatic striping, arbitrary cell 
formatting, automatic link creation, etc.

Supported data sets include arrayrefs of arrayrefs (as produced 
by the DBI selectall_arrayref, for example), arrayrefs of hashrefs, 
simple hashref (for a single row table), or DBIx::Recordset objects. 
By default arrayref-based datasets are interpreted as containing 
table rows; a column-based interpretation can be forced using 
style => 'across'.

The presentation definition can be defined in stages, if desired, 
and can handle arbitrary cell formatting, link creation, automatic 
striping, etc.


=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>


=head1 COPYRIGHT

Copyright 2002-2003, Gavin Carr and Open Fusion Pty Ltd. All Rights 
Reserved.

This program is free software. You may copy or redistribute it under the 
same terms as perl itself.

=cut
